# -*- coding: utf-8 -*-
"""
This module handles the extraction of the measured quantities in accordance
with the **ISO 13694** and **ISO 11145**.
"""

# =============================================================================
# Imports
# =============================================================================
import numpy as np


def max_power_density(raw_data):
    """
    `max_power_density` returns the maximum power density in the power density
    distribution. In order to avoid errors induced by background noise, it is
    recommended to use as input a noise-corrected dataframe.

    Parameters
    ----------
    raw_data : dataframe
        noise-corrected power density distribution.

    Returns
    -------
    float64
        maximum power density of the power density distribution.

    """
    # The first .max() returns the maximum value per each column, while the
    # second .max() returns the maximum value amount the maximums
    return raw_data.max().max()


def total_power(raw_data):
    """
    `total_power` returns the total power in the power density distribution. In
    order to avoid errors induced by background noise, it is recommended to
    use as input a noise-corrected dataframe.

    Parameters
    ----------
    raw_data : dataframe
        noise-corrected power density distribution distribution.

    Returns
    -------
    float64
        total power of the power density distribution.

    """

    # Convert dataframe to numpy array
    raw_data_np = raw_data.to_numpy()

    return np.nansum(raw_data_np)


def clip_level_power_density(raw_data, clip_level):
    """
    `clip_level_power_density` returns a fraction of the maximum power density.
    In order to avoid errors induced by background noise, it is recommended to
    use as input a noise-corrected dataframe.

    Parameters
    ----------
    raw_data : dataframe
        noise-corrected power density distribution.
    clip_level : float
        0 <= clip_level <= 1.

    Returns
    -------
    float64
        fraction of the maximum power density.

    """

    return max_power_density(raw_data) * clip_level


def clip_level_power(raw_data, clip_level):
    """
    `clip_level_power` returns the total power in the filtered power density
    distribution. Only power densities that are greater than the clip-level
    power density are considered in the calculation. In order to avoid errors
    induced by background noise, it is recommended to use as input a
    noise-corrected dataframe.

    Parameters
    ----------
    raw_data : dataframe
        noise-corrected power density distribution.
    clip_level : float
        0 <= clip_level <= 1. The clip-level defines the clip-level power
        density, and only power densities greater than to this threshold value
        are considered in the calculation.

    Returns
    -------
    float64
        total power considering only power densities that are greater than a
        threshold.

    """
    # The threshold is defined as the clip-level power density
    threshold = clip_level_power_density(raw_data, clip_level)
    raw_data_np = raw_data[raw_data >= threshold]

    return np.nansum(raw_data_np)



# -*- coding: utf-8 -*-
"""
This module handles the report generation of the beam analysis.
"""

# =============================================================================
# Imports
# =============================================================================
import os
import xlsxwriter


def write(path, fileName, beam):
    """
    `write` writes the beam analysis report in a `.xlsx` format.

    Parameters
    ----------
    path : str
        directory where the input file is saved.
    fileName : str
        name of the input file.
    beam : Beam
        object of type `Beam`.

    Returns
    -------
    None.

    """

    # Name of the input and output file
    fileName = os.path.splitext(fileName)[0]
    outFile = 'Beam Analysis - ' + fileName + '.xlsx'

    # Create a new Excel file
    wb = xlsxwriter.Workbook(os.path.join(path, outFile))

    # Create formats for this workbook
    bold = wb.add_format({
        'bold': True
    })
    italic = wb.add_format({
        'italic': True
    })
    number = wb.add_format({
        'num_format': '#,##0.00'
    })

    # Define a method to write the header
    def header(sheet):

        sheet.write('A1', 'Item', bold)
        sheet.write('B1', 'Value', bold)
        sheet.write('C1', 'Unit', bold)
        sheet.write('D1', 'Remark', bold)

    # Define a method to write the sub header
    def sub_header(sheet, row, value):

        sheet.write('A'+str(row), value, italic)

    # Define a method to write an empty line
    def empty_line(sheet, row):

        sheet.write('A'+str(row), '')

    # Define a method to write the entry
    def entry(sheet, row, value):

        sheet.write('A'+str(row), value[0])
        sheet.write('B'+str(row), value[1], number)
        sheet.write('C'+str(row), value[2])
        sheet.write('D'+str(row), value[3])

    # Define a method to insert an image
    def image(sheet, cell, path):

        sheet.insert_image(cell, path)

    def column_width(sheet):

        sheet.set_column('A:A', 30)
        sheet.set_column('B:B', 15)
        sheet.set_column('D:D', 70)

    # =========================================================================
    # Create and populate the ISO sheet
    # =========================================================================
    ws = wb.add_worksheet('ISO')
    column_width(ws)

    header(ws)

    sub_header(ws, 2, 'Beam power')
    entry(ws, 3, ['Total power',
                  beam.totalPower,
                  'ADC',
                  'Independent of the clip-level'])
    entry(ws, 4, ['Clip-level power',
                  beam.power_eta,
                  'ADC',
                  'Clip-level: ' + str(beam.eta*100) + '%'])
    empty_line(ws, 5)

    sub_header(ws, 6, 'Beam power density')
    entry(ws, 7, ['Maximum power density',
                  beam.maxPowerDensity,
                  'ADC',
                  'Independent of the clip-level'])
    entry(ws, 8, ['Clip-level power density',
                  beam.powerDensity_eta,
                  'ADC',
                  'Clip-level: ' + str(beam.eta*100) + '%'])
    entry(ws, 9, ['Clip-level average power density',
                  beam.averagePowerDensity_eta,
                  'ADC',
                  'Clip-level: ' + str(beam.eta*100) + '%'])
    empty_line(ws, 10)

    sub_header(ws, 11, 'Beam position')
    entry(ws, 12, ['Beam centroid x-axis',
                   beam.centerX * beam.xResolution,
                   'mm',
                   'Independent of the clip-level'])
    entry(ws, 13, ['Beam centroid y-axis',
                   beam.centerY * beam.yResolution,
                   'mm',
                   'Independent of the clip-level'])
    empty_line(ws, 14)

    sub_header(ws, 15, 'Effective beam size')
    entry(ws, 16, ['Beam width x-axis',
                   beam.widthX * beam.xResolution,
                   'mm',
                   'Independent of the clip-level'])
    entry(ws, 17, ['Beam width y-axis',
                   beam.widthY * beam.yResolution,
                   'mm',
                   'Independent of the clip-level'])
    entry(ws, 18, ['Clip-level irradiation area',
                   beam.irradiationArea_epsilon,
                   'mm',
                   'Clip-level: ' + str(beam.epsilon*100) + '%'])
    entry(ws, 19, ['Clip-level irradiation area',
                   beam.irradiationArea_eta,
                   'mm',
                   'Clip-level: ' + str(beam.eta*100) + '%'])
    empty_line(ws, 20)

    sub_header(ws, 21, 'Beam shape')
    entry(ws, 22, ['Beam aspect ratio',
                   beam.aspectRatio,
                   'N/A',
                   'Independent of the clip-level. Equals 1 for a perfect '
                   'square'])
    entry(ws, 23, ['Fractional power',
                   beam.fractionalPower_eta,
                   'N/A',
                   'Clip-level: ' + str(beam.eta*100) + '%.'])
    entry(ws, 24, ['Flatness factor',
                   beam.flatnessFactor_eta,
                   'N/A',
                   'Clip-level: ' + str(beam.eta*100) + '%. Equals 1 for a '
                   'perfect flat top'])
    entry(ws, 25, ['Beam uniformity',
                   beam.beamUniformity_eta,
                   'N/A',
                   'Clip-level: ' + str(beam.eta*100) + '%. Equals 0 for a '
                   'perfect flat top with vertical edges'])
    entry(ws, 26, ['Plateau uniformity',
                   beam.plateauUniformity_eta,
                   'N/A',
                   'Clip-level: ' + str(beam.eta*100) + '%. Equals 0 for a '
                   'perfect flat top'])
    entry(ws, 27, ['Edge steepness',
                   beam.edgeSteepness_eta,
                   'N/A',
                   'Clip-level: ' + str(beam.eta*100) + '%. Equals 0 for a '
                   'perfect vertical edge'])

    try:
        image(ws, 'E1', os.path.join(path, fileName + ' - histogram.png'))
        image(ws, 'E17', os.path.join(path, fileName + ' - 2d heat map.png'))
        image(ws, 'J17', os.path.join(path, fileName + ' - 3d heat map.png'))

    except xlsxwriter.exceptions.UnsupportedImageFormat:
        print("Unknown or unsupported image file format. Image will not be "
              "included in the summary file.")

    # =========================================================================
    # Create and populate the non ISO sheet
    # =========================================================================
    ws = wb.add_worksheet('non ISO')
    column_width(ws)

    header(ws)

    sub_header(ws, 2, 'Effective beam size')
    entry(ws, 3, ['Clip-level beam width x-axis',
                  beam.widthX_eta * beam.xResolution,
                  'mm',
                  'Clip-level: ' + str(beam.eta*100) + '%'])
    entry(ws, 4, ['Clip-level beam width y-axis',
                  beam.widthY_eta * beam.yResolution,
                  'mm',
                  'Clip-level: ' + str(beam.eta*100) + '%'])
    entry(ws, 5, ['Clip-level edge width x-axis',
                  beam.edgeX_epsilon_eta * beam.xResolution,
                  'mm',
                  'Clip-level: ' + str(beam.epsilon*100) + '% and '
                  + str(beam.eta*100) + '%'])
    entry(ws, 6, ['Clip-level edge width y-axis',
                  beam.edgeY_epsilon_eta * beam.yResolution,
                  'mm',
                  'Clip-level: ' + str(beam.epsilon*100) + '% and '
                  + str(beam.eta*100) + '%'])
    empty_line(ws, 7)

    sub_header(ws, 8, 'Beam shape')
    entry(ws, 9, ['Plateau uniformity',
                  beam.modPlateauUniformity_eta,
                  'N/A',
                  'Clip-level: ' + str(beam.eta*100) + '%. Equals 0 for a '
                  'perfect flat top'])
    entry(ws, 10, ['Top-hat factor',
                   beam.topHatFactor,
                   'N/A',
                   'Independent of the clip-level. Equals 1 for a perfect '
                   'square'])

    image(ws, 'E1', os.path.join(path, fileName + ' - energy curve.png'))

    try:
        wb.close()
    except xlsxwriter.exceptions.FileCreateError:
        print("The file is currently open and won't be saved. Please, close "
              "the file and run the analysis again.")



# -*- coding: utf-8 -*-
"""
Test file for a square beam.
"""
# =============================================================================
# Imports
# =============================================================================
import unittest
import pkg_resources

from beamprofiler import beam


class TestSquare(unittest.TestCase):
    """Tests for a perfectly flat-top beam."""

    def setUp(self):
        """`setUp` sets up the test fixtures."""

        path = pkg_resources.resource_filename(__name__,
                                               "fixtures")
        fileName = 'square_beam.xls'
        eta = 0.8
        epsilon = 0.1
        mix = 1
        self.beam = (
            beam.Beam(path, fileName, eta, epsilon, mix)
        )

    def test_totalPower(self):
        """`test_totalPower` tests the total power."""

        self.assertAlmostEqual(self.beam.totalPower, 10201000)

    def test_power_eta(self):
        """`test_power_eta` tests the clip-level power."""

        self.assertAlmostEqual(self.beam.power_eta, 10201000.0)

    def test_maxPowerDensity(self):
        """`test_maxPowerDensity` tests the maximum power density."""

        self.assertAlmostEqual(self.beam.maxPowerDensity, 1000)

    def test_powerDensity_eta(self):
        """`test_powerDensity_eta` tests the clip-level power density."""

        self.assertAlmostEqual(self.beam.powerDensity_eta, 800.0)

    def test_averagePowerDensity_eta(self):
        """`test_averagePowerDensity_eta` tests the clip-level average power
        density."""

        self.assertAlmostEqual(self.beam.averagePowerDensity_eta, 1000)

    def test_centerX(self):
        """`test_centerX` tests the center coordinate on the x-axis."""

        self.assertAlmostEqual(self.beam.centerX, 128)

    def test_centerY(self):
        """`test_centerY` tests the center coordinate on the y-axis."""

        self.assertAlmostEqual(self.beam.centerY, 128)

    def test_widthX(self):
        """`test_widthX` tests the beam width about the x-axis."""

        self.assertAlmostEqual(self.beam.widthX, 116.619)

    def test_widthY(self):
        """`test_widthY` tests the beam width about the y-axis."""

        self.assertAlmostEqual(self.beam.widthY, 116.619)

    def test_irradiationArea_epsilon(self):
        """`test_irradiationArea_epsilon` tests the lower clip-level
        irradiation area."""

        self.assertAlmostEqual(self.beam.irradiationArea_epsilon, 10201)

    def test_irradiationArea_eta(self):
        """`test_irradiationArea_eta` tests the upper clip-level irradiation
        area."""

        self.assertAlmostEqual(self.beam.irradiationArea_eta, 10201)

    def test_aspectRatio(self):
        """`test_aspectRatio` tests the beam aspect ratio."""

        self.assertAlmostEqual(self.beam.aspectRatio, 1.0)

    def test_fractionalPower_eta(self):
        """`test_fractionalPower_eta` tests the clip-level fractional power."""

        self.assertAlmostEqual(self.beam.fractionalPower_eta, 1.0)

    def test_flatnessFactor_eta(self):
        """`test_flatnessFactor_eta` tests the clip-level flatness factor."""

        self.assertAlmostEqual(self.beam.flatnessFactor_eta, 1.0)

    def test_beamUniformity_eta(self):
        """`test_beamUniformity_eta` tests the clip-level beam uniformity."""

        self.assertAlmostEqual(self.beam.beamUniformity_eta, 0.0)

    def test_plateauUniformity_eta(self):
        """`test_plateauUniformity_eta` tests the clip-level plateau
        uniformity."""

        self.assertAlmostEqual(self.beam.plateauUniformity_eta, 0.0)

    def test_edgeSteepness_eta(self):
        """`test_edgeSteepness_eta` tests the clip-level edge steepness."""

        self.assertAlmostEqual(self.beam.edgeSteepness_eta, 0.0)

    def test_widthX_eta(self):
        """`test_widthX_eta` tests the clip-level beam width about the
        x-axis."""

        self.assertAlmostEqual(self.beam.widthX_eta, 101.0)

    def test_widthY_eta(self):
        """`test_widthY_eta` tests the clip-level beam width about the
        y-axis."""

        self.assertAlmostEqual(self.beam.widthY_eta, 101.0)

    def test_edgeX_epsilon_eta(self):
        """`test_edgeX_epsilon_eta` tests the clip-level edge width about
        the x-axis."""

        self.assertAlmostEqual(self.beam.edgeX_epsilon_eta, 0.0)

    def test_edgeY_epsilon_eta(self):
        """`test_edgeY_epsilon_eta` tests the clip-level edge width about
        the y-axis."""

        self.assertAlmostEqual(self.beam.edgeY_epsilon_eta, 0.0)

    def test_modPlateauUniformity_eta(self):
        """`ModPlateauUniformity_eta` tests the modified clip-level plateau
        uniformity."""

        self.assertAlmostEqual(self.beam.modPlateauUniformity_eta, 0.0)

    def test_topHatFactor(self):
        """`test_topHatFactor` tests the top-hat factor."""

        self.assertAlmostEqual(self.beam.topHatFactor, 1.0)


if __name__ == '__main__':
    unittest.main()
